from __future__ import annotations
from enum import Enum
import re
from typing import List, Optional

from pydantic import BaseModel


class Matcher(BaseModel):
    major: List[re.Pattern] = []
    minor: List[re.Pattern] = []
    patch: List[re.Pattern] = []

    @staticmethod
    def default() -> Matcher:
        raw = {
            "major": [
                r"^(\* ?)?(hot)?fix ?(\(( ?\w)+\))?!: ",
                r"^(\* ?)?feat(ure)? ?(\(( ?\w)+\))?!: ",
                r"^(\* ?)?refactor(ing)? ?(\(( ?\w)+\))?!: ",
                r"(?i)^(\* ?)?BREAKING(?:\s*CHANGE)? ?(\(( ?\w)+\))?: ",
            ],
            "minor": [
                r"^(\* ?)?feat(ure)? ?(\(( ?\w)+\))?: ",
            ],
            "patch": [
                r"^(\* ?)?(hot)?fix ?(\(( ?\w)+\))?: ",
                r"^(\* ?)?refactor(ing)? ?(\(( ?\w)+\))?: ",
            ],
        }
        return Matcher(**raw)  # type: ignore


class Commit(BaseModel):
    username: Optional[str] = None
    email: Optional[str] = None
    message: str = "Version bumped from $version to $new_version"


class Type(str, Enum):
    GIT = "git"
    HG = "hg"
    SVN = "svn"


class VersionControl(BaseModel):
    type: Type = Type.GIT
    matcher: Matcher = Matcher.default()
    commit: Commit = Commit()
    tag: str = "v$new_version"



import copy
import os
from typing import Tuple
import unittest

from src.verbm.source import SourceManager
from src.verbm.version import Version
from src.verbm.config.config import Config


class TestSource(unittest.TestCase):
    def test_consistent(self):
        def prepare(filename: str) -> Tuple[SourceManager, Version]:
            c = Config.from_file(filename)
            sm = SourceManager(c.path, c.source)
            v = Version(c.template, c.version)

            return sm, v

        sm, v = prepare("./tests/data/source/consistent/version.yaml")
        self.assertTrue(sm.consistent(v))

        sm, v = prepare("./tests/data/source/inconsistent/version.yaml")
        self.assertFalse(sm.consistent(v))

        sm, v = prepare("./tests/data/source/invalid/version.yaml")
        self.assertRaises(Exception, sm.consistent, v)

    def test_replace(self):
        c = Config.from_file("./tests/data/source/consistent/version.yaml")
        sm = SourceManager(c.path, c.source)

        v1 = Version(c.template, c.version)
        self.assertTrue(sm.consistent(v1))

        v2 = copy.copy(v1)
        v2.patch += 1

        sm.replace(v1, v2)
        self.assertTrue(sm.consistent(v2))

        sm.replace(v2, v1)  # revert back
        self.assertTrue(sm.consistent(v1))

    def test_files(self):
        c = Config.from_file("./tests/data/source/consistent/version.yaml")
        sm = SourceManager(c.path, c.source)

        for f in sm.files():
            self.assertTrue(os.path.exists(f))


if __name__ == "__main__":
    unittest.main()



import os
import unittest
from src.verbm.init import init_project
from src.verbm.version_control.git import Git


class TestInit(unittest.TestCase):

    def test_init(self):
        git = Git()

        init_project("./tests/data", git)

        self.assertRaises(Exception, init_project, "./tests/data", git)

        os.remove("./tests/data/version.yml")


if __name__ == "__main__":
    unittest.main()



import os
import unittest

from pydantic import ValidationError
from src.verbm.config.config import Config


class TestConfig(unittest.TestCase):
    def test_config(self):
        root = "./tests/data/config"

        valid_dir = root + "/valid"
        filenames = next(os.walk(valid_dir), (None, None, []))[2]

        for filename in filenames:
            # doesn't raise
            _ = Config.from_file(valid_dir + "/" + filename)

        self.assertRaises(Exception, Config.from_file)

        self.assertRaisesRegex(
            ValidationError,
            "ver",
            Config.from_file,
            root + "/invalid/version.yml",
        )
        self.assertRaisesRegex(
            ValidationError,
            "template",
            Config.from_file,
            root + "/invalid/template.yml",
        )


if __name__ == "__main__":
    unittest.main()



from __future__ import annotations

from typing import List
import os

from pydantic import BaseModel
import yaml

from .version_control import VersionControl


DEFAULT_FILENAMES = ["version.yml", "version.yaml"]


class Source(BaseModel):
    file: str
    template: str


class Config(BaseModel):
    """
    Config is a root container of all configuration data, responsible for
    parsing the yaml file and storing data from it in a structured way.
    """

    path: str
    version: str
    template: str
    source: List[Source] = []
    version_control: VersionControl = VersionControl()

    @staticmethod
    def from_file(path: str) -> Config:
        path = Config.__path_or_default(path)

        with open(path, "r") as f:
            data = yaml.safe_load(f)
            data["path"] = path  # required field
            c = Config(**data)
            return c

        raise Exception(f"cannot load {path}")

    @staticmethod
    def __path_or_default(path: str | None) -> str:
        if path is not None:
            if os.path.isfile(path):
                return path

            raise Exception(f"no such file: '{path}'")

        # else
        for fn in DEFAULT_FILENAMES:
            if os.path.isfile(fn):
                return fn

        raise Exception(f"cannot find any of: {DEFAULT_FILENAMES}")



import os
import re
from string import Template
from typing import List

from .config.config import Source
from .version import Version


class SourceManager:
    """
    SourceManager is responsible for updating versions across source files.
    It keeps track of the absolute path to each file, including the configuration file, and can return these paths when needed.
    """

    cfg_path: str
    root: str
    sources: List[Source]

    def __init__(self, cfg_path: str, sources: List[Source]):
        self.cfg_path = os.path.realpath(cfg_path)

        root = os.path.dirname(self.cfg_path)
        self.root = root

        self.sources = sources
        for s in self.sources:
            s.file = os.path.realpath(os.path.join(self.root, s.file))

    def consistent(self, version: Version) -> bool:
        if not self.__contains(self.cfg_path, version):
            return False

        for src in self.sources:
            if not self.__contains(src.file, version):
                return False

        return True

    def __contains(self, path: str, version: Version) -> bool:
        """
        Checks the file with corresponding path contains specified version
        """
        if not os.path.isfile(path):
            raise Exception(f"cannot find: {path}")

        with open(path, "r+") as file:
            if file.read().find(str(version)) < 1:
                print(f"file: {path} doesn't contain version: {version}")
                return False

        return True

    def replace(self, old_version: Version, new_version: Version):
        with open(self.cfg_path, "r+") as file:
            content = file.read()

            # space tolerant before and after the colon
            regex = re.compile(f"version *:[ \n]*{re.escape(str(old_version))}")

            m = re.search(regex, content)
            if not m:
                raise Exception("cannot match version in the config file")

            old_str = m.group(0)
            new_str = old_str.replace(str(old_version), str(new_version))

            content = content.replace(old_str, new_str)

            file.seek(0)
            file.truncate()
            file.write(content)

        for src in self.sources:
            if not os.path.isfile(src.file):
                raise Exception(f"cannot find: {src.file}")

            old_str = Template(src.template).substitute(version=str(old_version))
            new_str = Template(src.template).substitute(version=str(new_version))

            with open(src.file, "r+") as file:
                content = file.read()
                content = content.replace(old_str, new_str)

                file.seek(0)
                file.truncate()
                file.write(content)

    def files(self) -> List[str]:
        return [self.cfg_path] + list(map(lambda x: x.file, self.sources))



from os import path
import re
from typing import Tuple
from .version_control.interface import VersionControl
from .config.config import DEFAULT_FILENAMES


def __version(vc: VersionControl) -> Tuple[str, str, str]:
    tm = re.compile("v?([0-9]+)\.([0-9]+)\.([0-9]+)")

    if (t := vc.last_tag()) is not None:
        m = re.match(tm, t)
        if m:
            return (m.group(1), m.group(2), m.group(3))
        else:
            print("cannot parse the last tag, use a default")
    else:
        print("cannot get the last tag, use a default")

    return ("0", "0", "0")


def init_project(dir: str, vc: VersionControl):
    # check existing of the config file first
    for fn in DEFAULT_FILENAMES:
        if path.exists(path.join(dir, fn)):
            raise Exception(
                f"directory '{dir}' contains {fn}, the project is already initialized"
            )

    major, minor, patch = __version(vc)

    username = vc.username()
    if not username:
        print("cannot get the username, use a placeholder")
        username = "John Doe"

    email = vc.email()
    if not email:
        print("cannot get the email, use a placeholder")
        email = "john.doe@example.com"

    yaml = f"""# main field, required
version: {major}.{minor}.{patch}

# template is needed for validation and splitting into components
# $major   integer, required
# $minor   integer, recommended
# $patch   integer, recommended
# $suffix  string,  optional
template: $major.$minor.$patch$suffix

# a list of source code files and corresponding patterns for version updates
# use `$version` as a placeholder
source:
  # - file: ./main.txt
  #   template: VERSION = "$version"


version_control:
  # default and single type, meaningless at this point
  # type: git

  commit:
    # the committer's username and email if different from the commit author from `git config`
    username: {username}
    email: {email}

    # commit message template, use `$version` and `$new_version` as placeholders, optional
    message: Version bumped from $version to $new_version

  # supports `$version` and `$new_version` placeholders the same as commit message, optional
  tag: v$new_version

  # default regex matchers for different version componets, optional
  matcher:
    major:
      - '^(\* ?)?(hot)?fix ?(\(( ?\w)+\))?!: '
      - '^(\* ?)?feat(ure)? ?(\(( ?\w)+\))?!: '
      - '^(\* ?)?refactor(ing)? ?(\(( ?\w)+\))?!: '
      - '(?i)^(\* ?)?BREAKING(?:\s*CHANGE)? ?(\(( ?\w)+\))?: '
    minor:
      - '^(\* ?)?feat(ure)? ?(\(( ?\w)+\))?: '
    patch:
      - '^(\* ?)?(hot)?fix ?(\(( ?\w)+\))?: '
      - '^(\* ?)?refactor(ing)? ?(\(( ?\w)+\))?: '
"""

    with open(path.join(dir, DEFAULT_FILENAMES[0]), "w") as file:
        file.write(yaml)



import copy
import re
from string import Template
from typing import List, Optional

from .clap import parser
from .config.config import Config
from .config.version_control import Matcher
from .version import Version
from .source import SourceManager
from .version_control.git import Git
from .version_control.interface import VersionControl
from .init import init_project

VERSION = "1.0.0"


def __to_component(commits: List[str], matcher: Matcher) -> Optional[str]:
    print("Commits to analyse:\n")
    for commit in commits:
        print(commit)

    for commit in commits:
        for r in matcher.major:
            if re.match(r, commit):
                return "major"

    for commit in commits:
        for r in matcher.minor:
            if re.match(r, commit):
                return "minor"

    for commit in commits:
        for r in matcher.patch:
            if re.match(r, commit):
                return "patch"

    return None


def run():
    # parse command line arguments
    args = parser().parse_args()

    vc = Git()
    if args.command == "init":
        init_project(".", vc)

    if args.version:
        print(VERSION)
        exit(0)

    cfg = Config.from_file(args.file)

    v = Version(format=cfg.template, version=cfg.version)

    if args.command == "get":
        print(v)
        exit(0)

    source = SourceManager(cfg.path, cfg.source)
    if not source.consistent(v):
        raise Exception("version is not consistent accross all files")

    if args.command == "validate":
        # already checked anyway, just print result
        print(f"version {v} is consistent accors all files")
        exit(0)

    # we are ready to check all other commands
    v2 = copy.copy(v)
    report = ""

    if args.command == "set":
        v2.parse(args.new_version)

        report = f"version was set to {v2}"

    elif args.command == "up":
        if args.component == "auto":
            filters = list(map(re.compile, args.filter))

            tag = Template(cfg.version_control.tag).substitute(new_version=str(v))

            component = __to_component(
                vc.log(tag, filters), cfg.version_control.matcher
            )
            if component == None:
                print("\nnothing to up")
                exit(0)
            else:
                print(f"\ncomponent to up: {component}")
                args.component = component

        if args.component == "major":
            v2.major += 1
            v2.minor, v2.patch = 0, 0
        elif args.component == "minor":
            v2.minor += 1
            v2.patch = 0
        elif args.component == "patch":
            v2.patch += 1

        report = f"version bumped to {v2}"

    elif args.command == "down":
        if args.component == "major":
            v2.major -= 1
        elif args.component == "minor":
            v2.minor -= 1
        elif args.component == "patch":
            v2.patch -= 1

        if v2.major < 0 or v2.minor < 0 or v2.patch < 0:
            raise Exception(f"version component cannot be negative {v2}")

        report = f"version downgraded to {v2}"

    source.replace(v, v2)
    print(report)

    # almost done, commit and push work
    if args.commit:
        vc.add(*source.files())

        report = Template(cfg.version_control.commit.message).substitute(
            version=str(v), new_version=str(v2)
        )
        vc.commit(
            report,
            username=cfg.version_control.commit.username,
            email=cfg.version_control.commit.email,
        )

    if args.tag:
        tag = Template(cfg.version_control.tag).substitute(new_version=str(v2))
        vc.tag(tag)

    if args.push:
        vc.push(args.tag)



from argparse import ArgumentParser


def __add_file(p: ArgumentParser):
    p.add_argument("-f", "--file", help="specify configuration file")


def __add_common(p: ArgumentParser):
    p.add_argument("-c", "--commit", help="commit changes", action="store_true")
    p.add_argument("-t", "--tag", help="add a tag with version", action="store_true")
    p.add_argument("-p", "--push", help="push changes", action="store_true")
    __add_file(p)


def parser() -> ArgumentParser:
    p = ArgumentParser(description="version manipulating tool")

    sub = p.add_subparsers(dest="command")

    # fmt: off
    p.add_argument(
        "-v", "--version",
        help="print verbm version",
        action="store_true",
        default=False
    )


    _ = sub.add_parser(
        "init",
        description=f"create a default configuration file",
    )


    pget = sub.add_parser(
        "get",
        description=f"print current version",
    )
    __add_file(pget)


    pset = sub.add_parser(
        "set",
        description="write version to the file"
    )
    pset.add_argument(
        "new_version",
        help="semantic version in <major.minor.patch> format",
        type=str
    )
    __add_common(pset)


    pval = sub.add_parser(
        "validate",
        description="check that all files contains the same version"
    )
    __add_file(pval)


    version_components = ["major", "minor", "patch"]

    pup = sub.add_parser(
        "up",
        description="up version"
    )
    pup.add_argument(
        "component",
        help="component of semantic version",
        type=str,
        choices=version_components+["auto"],
    )
    pup.add_argument(
        "-F", "--filter",
        nargs='+',
        type=str,
        default=[],
        help="optional regex file filter for a commit",
    )
    __add_common(pup)


    pdown = sub.add_parser(
        "down",
        description="down version"
    )
    pdown.add_argument(
        "component",
        help="component of semantic version",
        type=str,
        choices=version_components,
    )
    __add_common(pdown)
    # fmt: on

    return p



import re
from typing import List, Optional

from .interface import VersionControl
from .call import call


class Git(VersionControl):
    __delim = ">+>:+-\n"

    def add(self, *files: str):
        call("git", "add", *files)

    def commit(self, message: str, username: Optional[str], email: Optional[str]):
        committer = []
        if username:
            committer.extend(["-c", f'user.name="{username}"'])
        if email:
            committer.extend(["-c", f'user.email="{email}"'])

        call("git", *committer, "commit", "-m", message)

    def tag(self, tag: str):
        call("git", "tag", tag)

    def push(self, with_tags: bool):
        call("git", "push")

        if with_tags:
            call("git", "push", "--tags")

    def log(self, from_tag: str, file_filters: List[re.Pattern]) -> List[str]:
        # useful commands:
        # call("git", "rev-list", "--max-parents=0", "HEAD")

        hash = call("git", "rev-parse", from_tag).rstrip()

        out = call(
            "git",
            "--no-pager",
            "log",
            "--decorate=short",
            f"--pretty=format:{self.__delim}%s%n%b{self.__delim}",
            "--name-only",
            f"{hash}^..HEAD",
        )
        return self.__filter(out, file_filters)

    def __filter(self, data: str, file_filters: List[re.Pattern]) -> List[str]:
        # parse data to sequence ["", commit, files, commit, files, ...]
        chunks = data.split(self.__delim)

        commits: List[str] = []

        for i in range(2, len(chunks), 2):
            fnames = chunks[i].split("\n")[1:]  # starts from ""
            matched = False

            if len(file_filters) == 0:
                matched = True
            else:
                for fname in fnames:
                    for ffilter in file_filters:
                        matched |= re.match(ffilter, fname) != None

            if matched:
                # it could be squashed and contains a few commit messages
                commit = chunks[i - 1]
                lines = list(filter(lambda s: s != "", commit.split("\n")))
                commits.extend(lines)

        return commits

    def last_tag(self) -> Optional[str]:
        try:
            return call("git", "describe", "--tags", "--abbrev=0").rstrip()
        except:
            return None

    def username(self) -> Optional[str]:
        try:
            return call("git", "config", "user.name").rstrip()
        except:
            return None

    def email(self) -> Optional[str]:
        try:
            return call("git", "config", "user.email").rstrip()
        except:
            return None



from abc import ABC, abstractmethod
import re
from typing import List, Optional


class VersionControl(ABC):
    """
    An abstraction over git, mercurial and other version control systems
    """

    @abstractmethod
    def add(self, *files: str):
        pass

    @abstractmethod
    def commit(self, message: str, username: Optional[str], email: Optional[str]):
        pass

    @abstractmethod
    def tag(self, tag: str):
        pass

    @abstractmethod
    def push(self, with_tags: bool):
        pass

    @abstractmethod
    def log(self, from_tag: str, file_filters: List[re.Pattern]) -> List[str]:
        return []

    @abstractmethod
    def last_tag(self) -> Optional[str]:
        return None

    @abstractmethod
    def username(self) -> Optional[str]:
        return None

    @abstractmethod
    def email(self) -> Optional[str]:
        return None



from __future__ import annotations
import re
from string import Template
from typing import Optional


class Version:
    """
    Version encapsulate the semantic version splitted into its components.
    """

    major: int = 0
    minor: int = 0
    patch: int = 0
    suffix: Optional[str] = None

    __format: str
    __regex: re.Pattern

    def __init__(self, format: str, version: str):
        """
        Create a new Version object with the given format which is a template string
        with $major, $minor and $patch identifiers.

        If version is provided, it will be parsed and validated against the format.
        """

        regex = Template(format).substitute(
            major="([0-9]+)", minor="([0-9]+)?", patch="([0-9]+)?", suffix="(.+)?"
        )

        self.__format = format
        self.__regex = re.compile(f"^{regex}$")

        self.parse(version)

    def __str__(self) -> str:
        return Template(self.__format).substitute(
            major=self.major,
            minor=self.minor,
            patch=self.patch,
            suffix=self.suffix or "",
        )

    def parse(self, version: str):
        # validate first
        m = re.match(self.__regex, version)
        if not m:
            raise Exception(
                f"version: '{version}' and template: '{self.__format}' don't match each other"
            )

        # keep order of keywords
        placeholders = re.findall(r"\$(\w+)", self.__format)

        if len(m.groups()) != len(placeholders):
            raise Exception(f"not enough components in: {version}")

        # guarantee order
        parsed = {}
        for i, n in enumerate(m.groups()):
            parsed[placeholders[i]] = n

        self.major = int(parsed["major"])
        if "minor" in parsed:
            self.minor = int(parsed["minor"])
        if "patch" in parsed:
            self.patch = int(parsed["patch"])
        if "suffix" in parsed:
            self.suffix = parsed["suffix"]



import unittest
from src.verbm.version_control.call import call


class TestGit(unittest.TestCase):

    def test_call(self):
        out = call("ls", "-la")
        n = len(out.split("\n"))

        self.assertGreater(n, 1)

        # negative, wrong argument with "--"
        self.assertRaises(Exception, call, "ls", "--la")


if __name__ == "__main__":
    unittest.main()



from verbm.run import run


def main():
    try:
        run()
    except Exception as ex:
        # error only, without the stacktrace
        print(ex)
        exit(1)


if __name__ == "__main__":
    main()



import unittest
from src.verbm.version import Version


class TestVersion(unittest.TestCase):
    def test_init_parse(self):
        format = "v$major.$minor.$patch"
        version = "v1.2.3"

        # the first version
        v = Version(format, "v0.0.0")
        self.assertEqual(str(v), "v0.0.0")

        v.parse(version)

        self.assertEqual(v.major, 1)
        self.assertEqual(v.minor, 2)
        self.assertEqual(v.patch, 3)

        format = "v$major.$patch.$minor"  # wrong order
        v = Version(format, version)

        self.assertEqual(v.major, 1)
        self.assertEqual(v.patch, 2)
        self.assertEqual(v.minor, 3)

        self.assertRaises(Exception, v.parse, "3.4.5")  # without 'v'

    def test_components(self):
        v = Version("$major.$minor.$patch", "1.2.3")
        self.assertEqual(str(v), "1.2.3")

        v = Version("$major.$minor", "1.2")
        self.assertEqual(str(v), "1.2")

        v = Version("$major", "1")
        self.assertEqual(str(v), "1")

        # at least `major` component is required
        self.assertRaises(Exception, Version, "$major", "")
        self.assertRaises(Exception, Version, "$minor", "2")
        self.assertRaises(Exception, Version, "$patch", "3")

        # components are numbers
        self.assertRaises(Exception, Version, "$major.$minor.$patch", "foo.foo.foo")
        self.assertRaises(Exception, Version, "$major.$minor", "foo.foo")
        self.assertRaises(Exception, Version, "$major", "foo")

    def test_suffix(self):
        format = "v$major.$minor.$patch$suffix"
        version = "v1.2.3-rc"

        v = Version(format, version)
        self.assertEqual(str(v), "v1.2.3-rc")
        self.assertIsNotNone(v.suffix)
        self.assertEqual(v.suffix, "-rc")


if __name__ == "__main__":
    unittest.main()



import subprocess
from typing import IO, List, cast


def call(command: str, *args) -> str:
    """
    call command with arguments and return stdout or raise exception
    """

    p = subprocess.Popen(
        args=[command] + [a for a in args],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    p.wait()

    def read(stream: IO[bytes]) -> str:
        def read_one():
            data = stream.readline()
            if not data:
                return ""
            return data.decode("utf-8")

        return "".join([line for line in iter(read_one, "")])

    out: str
    err: str

    if p.stdout:
        out = read(p.stdout)
        p.stdout.close()

    if p.stderr:
        err = read(p.stderr)
        p.stderr.close()

    if p.returncode != 0:
        if err:
            raise Exception(err)
        else:
            cmd = " ".join(cast(List[str], p.args))
            raise Exception(f"'{cmd}'\n failed with exit code: {p.returncode}")

    return out
